unit model;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,IniFiles,RTTIUtils;

Type

  TEntity = Class(TComponent)
  private
    FID : String;
    FModified: Boolean;
    FIni : TInifile;
  Protected
    Procedure GetPropertyList(L : TStrings);
    Function  CreatePropStorage : TPropsStorage;
    function  ReadStr(const ASection, Item, ADefault: string): string;
    procedure WriteStr(const ASection, Item, Value: string);
    procedure EraseSection(const ASection: string);
    Procedure SetModified;
    procedure LoadFromStorage(AID : String);
    procedure SaveToStorage;
  Public
    Destructor Destroy; override;
    Property ID : String Read FID;
    Property Modified : Boolean Read FModified;
  end;

  TModel = Class;

  TDataAgent = Class(TObject)
  private
    FModel: TModel;
    FOwner: TComponent;
    procedure SetModel(const AValue: TModel);
  Public
    Constructor Create(AOwner : TComponent);
    Destructor Destroy; override;
    Procedure SetEntity(AEntity : TEntity); virtual;Abstract;
    Property Model : TModel Read FModel Write SetModel;
    Property Owner : TComponent Read FOwner;
  end;
  
  TModel = Class(TComponent)
  Private
    FAgentList : TList;
  Protected
    Procedure ClearAgents; virtual;
    Procedure NotifyAgents(AEntity : TEntity); virtual;
    function GetNewID(EntityName : String) : Integer;
  Public
    Constructor Create(AOwner : TComponent); override;
    Destructor Destroy; override;
    Function CreateNewInstance(EntityName : String) : TEntity;
    Function CreateInstanceFromID(EntityName,ID : String) : TEntity;
    Procedure RegisterDataAgent(Sender : TDataAgent);
    Procedure UnRegisterDataAgent(Sender : TDataAgent);
    
  end;
  
  TPerson = Class(TEntity)
  private
    FBirthDay: TDateTime;
    FFirstName: String;
    FLastName: String;
    procedure SetBirthDay(const AValue: TDateTime);
    procedure SetFirstName(const AValue: String);
    procedure SetLastName(const AValue: String);
  Published
    Property FirstName : String Read FFirstName Write SetFirstName;
    Property LastName : String Read FLastName Write SetLastName;
    Property BirthDay : TDateTime Read FBirthDay Write SetBirthDay;
  end;


Procedure Register;

implementation

{ TEntity }

uses typinfo,lresources;

Procedure Register;

begin
  RegisterComponents('MDA',[TModel]);
end;

procedure TEntity.LoadFromStorage(AID: String);

Var
  L : TStringList;
  PS : TPropsStorage;
  
begin
  FID:=AID;
  L:=TStringList.Create;
  Try
    GetPropertyList(L);
    PS:=CreatePropStorage;
    Try
      FIni:=TInifile.Create(ClassName+'.dat');
      Try
        PS.LoadProperties(L);
        FIni.UpdateFile;
      Finally
        FreeAndNil(FIni);
      end;
    Finally
      PS.Free;
    end;
  Finally
    L.Free;
  end;
end;

procedure TEntity.SaveToStorage;

Var
  L : TStringList;
  PS : TPropsStorage;

begin
  FID:=ID;
  L:=TStringList.Create;
  Try
    GetPropertyList(L);
    PS:=CreatePropStorage;
    Try
      FIni:=TInifile.Create(ClassName+'.dat');
      Try
        PS.StoreProperties(L);
        FIni.UpdateFile;
        FModified:=False;
      Finally
        FreeAndNil(FIni);
      end;
    Finally
      PS.Free;
    end;
  Finally
    L.Free;
  end;
end;

Destructor TEntity.Destroy;
begin
  if Modified then
    SaveToStorage;
  inherited Destroy;
end;

procedure TEntity.GetPropertyList(L: TStrings);

Var
  I : Integer;
  PIL : TPropInfoList;
  P : PPropInfo;

begin
  PIL:=TPropInfoList.Create(Self,tkProperties);
  Try
    For I:=0 to PIL.Count-1 do
      begin
      P:=PIL.Items[i];
      L.Add(P^.Name);
      end;
  Finally
    PIL.Free;
  end;
end;

function TEntity.CreatePropStorage: TPropsStorage;
begin
  Result:=TPropsStorage.Create;
  With Result do
    begin
    AObject:=Self;
    Section:=FID;
    OnReadString:=@ReadStr;
    OnWriteString:=@WriteStr;
    OnEraseSection:=@EraseSection;
    end;
end;

function TEntity.ReadStr(const ASection, Item, ADefault: string): string;
begin
  If Assigned(FIni) then
    Result:=FIni.ReadString(ASection,Item,ADefault);
end;

procedure TEntity.WriteStr(const ASection, Item, Value: string);
begin
  If Assigned(FIni) then
    Fini.WriteString(ASection,Item,Value);
end;

procedure TEntity.EraseSection(const ASection: string);
begin
  If Assigned(Fini) then
    FIni.EraseSection(ASection);
end;

procedure TEntity.SetModified;
begin
  FModified:=True;
end;

{ TPerson }

procedure TPerson.SetBirthDay(const AValue: TDateTime);
begin
  if FBirthDay<>AValue then
    begin
    FBirthDay:=AValue;
    SetModified;
    end;
end;

procedure TPerson.SetFirstName(const AValue: String);
begin
  if FFirstName<>AValue then
    begin
    FFirstName:=AValue;
    SetModified;
    end;
end;

procedure TPerson.SetLastName(const AValue: String);
begin
  if FLastName<>AValue then
    begin
    FLastName:=AValue;
    SetModified;
    end;
end;

{ TModel }

procedure TModel.RegisterDataAgent(Sender: TDataAgent);
begin
  If FAgentList.indexOf(Sender)=-1 then
    FAgentList.Add(Sender);
end;

procedure TModel.UnRegisterDataAgent(Sender: TDataAgent);

Var
  Index : Integer;

begin
  Index:=FAgentList.indexOf(Sender);
  If Index<>-1 then
    FAgentList.Delete(Index);
end;

procedure TModel.ClearAgents;

Var
  I : Integer;
  
begin
  For I:=FAgentList.Count-1 downto 0 do
    TDataAgent(FAgentList[i]).FModel:=Nil;
  FAgentList.Clear;
end;

procedure TModel.NotifyAgents(AEntity : Tentity);

Var
  I : Integer;

begin
  For I:=0 to FAgentList.Count-1 do
    begin
    TDataAgent(FAgentList[i]).SetEntity(AEntity);
    Writeln('Notifying agent ',i,' : ',AEntity.Id);
    end;
end;

constructor TModel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAgentList:=TList.Create;
end;

destructor TModel.Destroy;
begin
  ClearAgents;
  FAgentList.Free;
  inherited Destroy;
end;

function TModel.CreateNewInstance(EntityName: String): TEntity;
begin
  Result:=TPerson.Create(Self);
  Result.FID:='Person'+IntToStr(GetNewID('TPerson'));
  NotifyAgents(Result);
end;

function TModel.GetNewID(EntityName : String) : Integer;

Var
  Ini : TMemInifile;

begin
  Ini:=TMemIniFile.Create(EntityName+'.dat');
  With Ini do
    Try
      Result:=ReadInteger('Global','LastID',0)+1;
      WriteInteger('Global','LastID',Result);
      UpdateFile
    Finally
      Free;
    end;
end;


function TModel.CreateInstanceFromID(EntityName, ID: String): TEntity;
begin
  Result:=TPerson.Create(Self);
  Result.LoadFromStorage(ID);
  NotifyAgents(Result);
end;

{ TDataAgent }

procedure TDataAgent.SetModel(const AValue: TModel);
begin
  if FModel<>AValue then
    begin
    If Assigned(FModel) then
      FModel.UnRegisterDataAgent(Self);
    FModel:=AValue;
    FModel.RegisterDataAgent(Self);
    end;
end;

constructor TDataAgent.Create(AOwner: TComponent);
begin
  FOwner:=AOwner;
end;

destructor TDataAgent.Destroy;
begin
  if Assigned(FModel) then
    FModel.UnRegisterDataAgent(Self);
  FModel:=Nil;
  inherited Destroy;
end;

initialization
{$i model.lrs}
end.
